import { message, GLOBAL_CONFIG_FILE } from "bddx-core";
import fs from "fs";
import inquirer from "inquirer";
import { TEST_STATUS } from "./doTests.js";
import { Chain } from "./zeus/index.js";
const chain = (option, Key) => Chain("https://bddx-p-api.azurewebsites.net/graphql", {
    headers: {
        "Content-type": "application/json",
        Key,
    },
})(option);
const uploadReports = async (uploadReportInput, key) => {
    const response = await chain("mutation", key)({
        cli: {
            uploadReport: [
                {
                    uploadReportInput,
                },
                true,
            ],
        },
    });
    if (!response) {
        message("We cannot upload your reports to BDDX Cloud right now.", "red");
        return;
    }
    return response.cli?.uploadReport;
};
export const cloudIntegration = async (resultsPaths) => {
    const selected = await inquirer.prompt([
        {
            type: "list",
            name: "pathtoresult",
            choices: resultsPaths,
            message: "Select file to publish on cloud",
            default: GLOBAL_CONFIG_FILE.in,
        },
        {
            type: "input",
            name: "key",
            message: "Insert your project key from BDDX Cloud:",
        },
    ]);
    if (selected.key && selected.pathtoresult) {
        const content = JSON.parse(fs.readFileSync(selected.pathtoresult).toString("utf-8"));
        if (content.testStatus.status === TEST_STATUS.FINISHED) {
            const arrayToSend = { results: [] };
            content.passedTests.map((test) => {
                arrayToSend.results.push({
                    testPath: test.testPath,
                    featureContent: test.featureContent,
                    testContent: test.scenarioContent,
                    scenario: test.scenarioTitle,
                });
            });
            content.failedTests.map((test) => {
                arrayToSend.results.push({
                    testPath: test.testPath,
                    featureContent: test.featureContent,
                    scenario: test.scenarioTitle,
                    testContent: test.scenarioContent,
                    reasonOfFail: test.reasonOfFail,
                });
            });
            if (arrayToSend.results.length === 0) {
                message("No results to send.", "red");
                return;
            }
            if (!selected.key) {
                message("Missing BDDX Cloud key.", "red");
                return;
            }
            if (arrayToSend.results.length) {
                try {
                    const response = await uploadReports(arrayToSend, selected.key);
                    if (!response) {
                        message("Error occurred while uploading results.", "red");
                        return;
                    }
                    message(`Success upload results with id: ${response}.`, "blue");
                    return;
                }
                catch {
                    message("Cannot upload results now, try again.", "red");
                    return;
                }
            }
        }
        else {
            message("This results have status of unfinished.", "red");
            return;
        }
    }
    else {
        message("You provided wrong value into fields.", "red");
        return;
    }
};
//# sourceMappingURL=cloudIntegration.js.map